/******************************************************************************

  CEdgeQueue.C

  Stephen Freund
  freunds@cs.stanford.edu

  Implementation of CEdgeQueue.

  The value for DerivativeAcrossEdge is different for directed and undirected
  graphs.  For undirected, it is the absolute value of the potential function
  use in directed graphs.  RealDerivative will always return the actual
  value of the potential function.

  In attempt to reduce the number times the derivative of a function is
  taken, the derivative will not be recomputed every time- only when
  atleast one of the queues has changed by one sigma unit.  The
  following methods that change the queue size return TRUE when this
  is the case.  Do, when these functions return TRUE, recompute the 
  derivatives in the heap.
	
******************************************************************************/

#include "CEdgeQueue.h"
#include "CHeap.h"
#include "CCommodity.h"
#include "Utility.h"
#include <math.h>
#include "Parameter.h"

/*
 * Just init everything to 0.
 */
CEdgeQueue::CEdgeQueue(CCommodity *comm)  {
  change[head] = change[tail] = 0;
  itsLevel[head] = itsLevel[tail] = 0;
  itsCommodity = comm;
  itsFlow = 0.0;
}

/*
 * Print the information about an edgequeue in a sort-of readable way.
 */
ostream& operator<<(ostream& out,const CEdgeQueue& q)  {
  double unitSize = q.itsCommodity -> OriginalDemand() * GetSigma();
  out << "(comm" << q.itsCommodity->Id();
  out <<" t:" << q.itsLevel[tail] / unitSize <<  " h:";
  out << q.itsLevel[head] /unitSize << ": flow: " << q.itsFlow << ") ";
  return(out);
}


/*
 * 
 */
CEdgeQueue::CEdgeQueue()  {
  CEdgeQueue(NULL);
}


/* 
 * 
 */
void CEdgeQueue::SetCommodity(CCommodity *comm)  {
  itsCommodity = comm;
}

/*
 * return TRUE if the commodity level has changed significantly since
 * the last time the derivative was computed.
 */
int CEdgeQueue::SetCommodityLevel(EdgePos pos, double level)  {
  change[pos] += (level - itsLevel[pos]);
  itsLevel[pos] = level;
  return(fabs(change[pos]) > GetSigma() * itsCommodity -> OriginalDemand());
}


/*
 * Move the given amount from the tail to the head, and increment the
 * flow accordingly.  return TRUE if the derivative needs to be 
 * recomputed.
 */
int CEdgeQueue::MoveCommodityAcross(double amount)  {
  int hVal = SetCommodityLevel(head, itsLevel[head] + amount);
  int tVal = SetCommodityLevel(tail, itsLevel[tail] - amount);
  itsFlow += amount;
  return (tVal || hVal);
}

/*
 * Add the given amount to the specified queue.  Return TRUE if it is
 * time to recompute the derivative.
 */
int CEdgeQueue::AdjustCommodityLevel(EdgePos pos, double deltalevel) { 
  double newLevel = itsLevel[pos] + deltalevel;
  return(SetCommodityLevel(pos, newLevel));  
}


/*
 * Compute the derivative across the edge assuming the levels of the
 * queues if an additional adjustment were moved accross.
 */
double CEdgeQueue::DerivativeAcrossEdge(double transferAdjustment)  {
  double deriv = RealDerivative(transferAdjustment);

#ifdef UNDIRECTED  
  return fabs(deriv);
#else
 return deriv;
#endif
}


double CEdgeQueue::RealDerivative(double transferAdjustment)  {
  double alpha = GetAlpha();
  double d = itsCommodity -> OriginalDemand();
  double deriv =  alpha / d * 
	 (MyExp( alpha * (itsLevel[tail] - transferAdjustment - kDelta) / d) -
          MyExp( alpha * (itsLevel[head] + transferAdjustment + kDelta) / d));
  
  return (deriv);
}





