/******************************************************************************

  CNode.h

  Stephen Freund
  freunds@cs.stanford.edu

  A node keeps a list of CEdge objects for those edges which leave the 
  node.  Whenever an operation is performed on all edges, each node 
  performs it on the edges in this list.  A second list of edges 
  incident to this node is also kept to be used during balancing and
  such.  These list are maintained as linked lists of CNodeItems.

  Arrays are maintained of the maximum and mininum number of sigma
  units present in any queues at the node for each commodity.  If the
  difference is less than 2, then no balancing within the node is
  performed.  If the difference is 2, then each maxmimum queue (or as
  many as possible) is reduced by 1 unit, and as many minimum queues
  as possible are increased by 1 unit.  Actual lists of queues with the 
  largest and smallest heights are not kept but are computed as part
  of the balancing process.  This is due to the need for both insertion
  and deletion in a list if the queues are kept, making it possibly 
  slower if many queues are adjusted each iteration.

  If the difference is greater than two (the only time this happens is
  right at the beginning at the sources and sinks, and occasionally during 
  execution), the number of sigma units of a commodity is summed for all
  queues, and then divided evenly among the queues.  This is slow, but
  is only needed infrequently during execution.

  NOTE: You must call InitEdgeQueues after the graph has been constructed
        and before beginning the balancing.  This method will create the
	edqe queues and other data structures used by the node.

******************************************************************************/

#ifndef _H_CNode
#define	_H_CNode

#include <iostream.h>
#include "CEdge.h"

class CNode;
class CGraph;

/*
 * Simple linked list item to make a list of CEdges 
 */
class CNodeItem {
   friend class CNode;
 private:
   CEdge *itsEdge;
   CNodeItem *itsNext;
  
   CNodeItem(CEdge *, CNodeItem*);
  
 public:
   ~CNodeItem() { delete itsNext; }
  
   friend ostream& operator<<(ostream&, const CNodeItem&);
   friend ostream& operator<<(ostream&,const CNode&);
   friend istream& operator>>(istream&, CNode&);
};


class CNode {
 private:
   int itsId;
   CNodeItem *itsEdges;         /* linked list of exiting edges */
   CNodeItem *itsEnteringEdges; /* linked list of entering edges */
   int itsDegree;               /* total number of edges incident on node */
   CGraph *itsGraph; 
   int *maxValues;              /* array of unit maximums for each commodity */
   int *minValues;              /* array of unit minimums for each commodity */
   
  
   /* helpers in balancing process */
   void BalanceCommodityInNode(int commodityIndex);  
   void DistributeUnitsToEachNodeByUnits(int commodityIndex, int amount);
   int ComputeAmountOfCommodityByUnits(int index);
   void BalanceFor2Diff(int commodityIndex);
   
   /* not used anymore: balances exactly. */
   void DistributeUnitsToEachNode(int commodityIndex, double amount);  

 public:
  
   /* constructors / destructors */
   CNode();
   CNode(CGraph *theGraph, int id);
   ~CNode();
  
   /* routines for building */
   void AddEdge(CNode *dest) ;
   void AddEnteringEdge(CEdge *incidentEdge);
   void InitEdgeQueues(int numCommodities, CCommodity *comm[]);
   
   
   /* accessors */
   int Id(void)		{ return itsId; }  
   CEdge *GetEdge(CNode *dest) const;
   int GetDegree(void) const { return itsDegree; }
   CGraph * GetGraph(void)  { return itsGraph; }
   
   /* balancing */
   void BalanceWithinNode(void);
   void BalanceAcrossEdges(void);
   
   /* to notify min/max values of change in queue height */
   void ChangeSizeOfQueue(CEdge *theEdge, int com, int units);
   
   /* adjusting sizes after iteration or phase */
   void AdjustEdgeCapacities(void);
   
   void HalveEdgeCapacitiesAndQueues(void);
   void DivideEdgeCapacitiesAndQueues(double amount);
   
   /* computing largest / smallest excess (sigma * d units) */
   CEdge *FindSmallestQueueForOutEdge(int commodityIndex);
   CEdge *FindLargestQueueForInEdge(int commodityIndex);
   CEdge *FindSmallestQueueForInEdge(int commodityIndex);
   CEdge *FindLargestQueueForOutEdge(int commodityIndex);
   
   /* amount of commodity in queues (all or only positive) */
   double ComputeAmountOfCommodity(int index);
   double ComputePositiveAmountOfCommodity(int index);
   
   /* I/O */
   friend ostream& operator<<(ostream&,const CNode&);
   void PrintFlows(void);
};

#endif



